__all__ = ["Factory", "factory", "Base"]

from .base import (Annotated, Any, AsyncAttrs, AsyncEngine, AsyncSession, Base,
                   DBModel, BinaryExpression, BooleanClauseList, ClauseList,
                   ClausesType, ColumnProperty, Coroutine, DeclarativeBase,
                   Dict, Factory, ForeignKey, Generic, IllegalStateChangeError,
                   InstrumentedAttribute, Integer, List, Mapped,
                   ResultWithStats, Row, Select, Sequence, String, Task,
                   TimeStamp, Type, TypeDecorator, TypeVar, UnaryExpression,
                   all_, and_, any_, configure_factory, factory, func,
                   mapped_column, or_, relationship, select)
from .base import *
from .types import *
from sqlalchemy import *



import logging
from types import NotImplementedType
from typing import Type, List

import fastapi.security.oauth2
from fastapi import APIRouter, Depends
from fastapi.routing import APIRoute

from ._repository import BaseRepository
from .types import DBModel, ReadModel, UpdateModel, CreateModel, FilterParams, PaginatedApiResponse
from ..db.types import ResultWithStatsType

logger = logging.getLogger(__name__)


class BaseRouter(APIRouter):
    """
    A base "Router" or fastapi Router.
    by passing in (either):
     - a DB model (SQLAlchemy), and Pydantic Models for Read, Create, Update
     - a Repository object
    it will return a APIRouter derivative, which contains 4 main actions but is of course
    extensible and customizable. All neatly document in FastAPI/OpenAPI docs/spec.

    this router can (should) work together with :ref:Repository
    the returned object consists of 4 main actions:

        - get_items()
            returns a list of items, based on filter params
            route: GET "/"

        - get_item()
            returns a single item, based on id
             route: "GET" "/{item_id:str}"
        - create_item()
            creates a new item
            route: "POST" "/"

        - update_item()
            updates an existing item
            route: "PATCH" "/{item_id:str}"

        - delete()
            deletes an item based on id
            route: "DELETE" "/{item_id:str}"

    """
    def __init__(
            self,
            *,
            db_model: Type[DBModel],
            model_create: Type[ReadModel] = None,
            model_read: Type[CreateModel] = None,
            model_patch: Type[UpdateModel] = None,
            repo_object: BaseRepository = None,
            prefix="",
            tags: List[str] = None,
            dependencies: List = None ,
            security: fastapi.security.oauth2.SecurityBase = None,
            redirect_slashes: bool = True,
            lifespan=None,
            include_in_schema: bool = True
    ):
        """

        :param db_model: the sqlalchemy db model class
        :param model_read: the pydantic model for read(get) operations
        :param model_create: the pydantic model for create(post) operations
        :param model_patch:  the pydantic model for update (patch) operations
        :param repo_object:  a custom repository instance (in case custom logic is needed)
        :param prefix:
        :param tags:
        :param dependencies:
        :param security:
        :param redirect_slashes:
        :param lifespan:
        :param include_in_schema:
        """
        super().__init__(
            prefix=prefix,
            tags=tags,
            dependencies=dependencies or [],
            redirect_slashes=redirect_slashes,
            lifespan=lifespan,
            include_in_schema=include_in_schema,
        )
        self.model_read = model_read
        self.model_create = model_create
        self.model_patch = model_patch
        self.security = security
        self.dependencies = dependencies or []
        if repo_object:
            self.repository = repo_object
        else:
            self.repository = BaseRepository[
                db_model, model_read, model_create, model_patch
            ](db_model)
        self.create_default_routes()

    async def list_response_hook(self, response_dict: ResultWithStatsType[ReadModel]):
        """
        a hook which is called before returning the final response of a list call (paginated response)
        this is dirty hack to enable subclassing (since all routes are created within create_default_routes()
        which in turn is a shortcoming in FastApi, which cannot handle the self param.
        this way data can be modified or checked if needed
        """
        return NotImplemented

    async def post_create_hook(self, data: CreateModel) -> CreateModel:
        """
        a hook which is called before further data processing.
        this is dirty hack to enable subclassing (since all routes are created within create_default_routes()
        which in turn is a shortcoming in FastApi, which cannot handle the self param.
        this way data can be modified or checked if needed
        """
        return NotImplemented

    async def get_read_hook(self, response: ReadModel) -> ReadModel:
        """
        a hook which is called before returning the final response of a list call (paginated response)
        this is dirty hack to enable subclassing (since all routes are created within create_default_routes()
        which in turn is a shortcoming in FastApi, which cannot handle the self param.
        this way data can be modified or checked if needed
        :param response:
        :type response:
        :return:
        :rtype:
        """
        return NotImplemented

    async def patch_update_hook(self, data: UpdateModel) -> UpdateModel:
        """
        a hook which is called before further data processing.
        this is dirty hack to enable subclassing (since all routes are created within create_default_routes()
        which in turn is a shortcoming in FastApi, which cannot handle the self param.
        this way data can be modified or checked if needed
        """
        return NotImplemented

    def create_default_routes(self):

        # these 3 are the response models passed in at __init__, and now used to annotate the return type of
        # our routes
        ModelRead = self.model_read  # noqa
        ModelCreate = self.model_create  # noqa
        ModelPatch = self.model_patch  # noqa

        self.routes.clear()

        extra = {}
        deps = []
        if self.security:
            self.dependencies.append(self.security)
            # extra.update(dependencies=[Depends(self.security)])
            # extra.update(dependencies=extra["dependencies"] + [*self.dependencies])

        self.dependencies = [Depends(x) if not hasattr(x, 'dependency') else x for x in self.dependencies]

        async def get_items(
                filter_params: FilterParams = Depends(FilterParams)
        ) -> PaginatedApiResponse[ModelRead]:
            r: ResultWithStatsType[ReadModel] = await self.repository.list(filter_params)

            x = await self.list_response_hook(r)
            if not isinstance(x, NotImplementedType):
                return x
            return r

        async def get_item(
                item_id: str,
        ) -> ModelRead:
            r = await self.repository.get(item_id)
            x = await self.get_read_hook(r)
            if not isinstance(x, NotImplementedType):
                return x
            return r

        async def create_item(
                data: ModelCreate
        ) -> ModelRead:
            x = await self.post_create_hook(data)
            if not isinstance(x, NotImplementedType):
                data = x
            return await self.repository.create(data)

        async def update_item(
                item_id: str, data: ModelPatch
        ) -> ModelRead:
            x = await self.patch_update_hook(data)
            if not isinstance(x, NotImplementedType):
                data = x
            return await self.repository.update(item_id, data)

        async def delete_item(item_id: str) -> str:
            return await self.repository.delete(item_id)

        self.add_api_route("/", get_items, methods=["GET"])
        self.add_api_route("/{item_id:str}", get_item, methods=["GET"])
        self.add_api_route("/", create_item, methods=["POST"])
        self.add_api_route("/{item_id:str}", update_item, methods=["PATCH"])
        self.add_api_route("/{item_id:str}", delete_item, methods=["DELETE"])
        return self

    def replace_route(self, index: int, new_route: APIRoute):
        self.routes[index] = new_route
        return self



import logging.config
import pathlib
from typing import Mapping as _Mapping
from typing import Sequence as _Sequence

import yaml

logger = logging.getLogger(__name__)


class Configuration(dict):

    def __init__(self, *args, **kwargs):
        super().__init__()
        super().__setattr__("__dict__", self)
        for k, v in dict(*args, **kwargs).items():
            super().__setattr__(k, self.__wrap(v))

    def __setattr__(self, key, value):
        super().__setattr__(key, self.__wrap(value))

    def __setitem__(self, key, value):
        super().__setattr__(key, self.__wrap(value))

    @classmethod
    def __wrap(cls, obj):
        if isinstance(obj, _Mapping):
            return cls(obj)
        if isinstance(obj, _Sequence) and \
                not isinstance(obj, str):
            return [cls.__wrap(o) for o in obj]
        return obj


def load_config(config: pathlib.Path | dict | str) -> Configuration:
    if isinstance(config, str):
        try:
            config = yaml.safe_load(open(config).read())
        except:
            logger.debug(f"failed to load configuration file from string '{config}'", exc_info=True)
    elif isinstance(config, dict):
        config = config
    elif isinstance(config, pathlib.Path):
        config = yaml.safe_load(config.read_text("r", encoding="utf8"))

    return Configuration(config)


def configure_logging():
    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {'colored': {'()': 'fastapis.common.loggers.CustomFormatter',
                                   'datefmt': '%(asctime)s'},
                       'json': {'()': 'fastapis.common.loggers.CustomJSONFormatter',
                                'datefmt': '%(asctime)s'}},
        'handlers': {'console': {'level': 20,
                                 'formatter': 'colored',
                                 'class': 'logging.StreamHandler',
                                 'stream': 'ext://sys.stdout'},
                     'file': {'level': 10,
                              'formatter': 'json',
                              'class': 'logging.handlers.RotatingFileHandler',
                              'filename': 'app.log',
                              'maxBytes': 1048576,
                              'backupCount': 3}},
        'loggers': {'root': {'handlers': ['console'],
                             'level': 'INFO',
                             'propagate': False},
                    'uvicorn': {'handlers': ['console', 'file'],
                                'level': 'DEBUG',
                                'propagate': False},
                    'uvicorn.access': {'handlers': ['console', 'file'],
                                       'level': 'DEBUG',
                                       'propagate': False},
                    'uvicorn.error': {'handlers': ['console', 'file'],
                                      'level': 'DEBUG',
                                      'propagate': False},
                    'uvicorn.asgi': {'handlers': ['console'],
                                     'level': 'DEBUG',
                                     'propagate': False}}}
    logging.config.dictConfig(config)



import json
import logging
import logging.config


class CustomFormatter(logging.Formatter):
    red = "\x1b[31m"
    red_dimmed = "\x1b[2;31m"
    red_bold = "\x1b[1;31m"

    yellow = "\x1b[32m"
    yellow_dimmed = "\x1b[2;32m"
    yellow_bold = "\x1b[1;32m"

    blue = "\x1b[34m"
    blue_dimmed = "\x1b[2;34m"
    blue_bold = "\x1b[1;34m"

    magenta = "\x1b[35m"
    magenta_dimmed = "\x1b[2;35m"
    magenta_bold = "\x1b[1;35m"

    cyan = "\x1b[36m"
    cyan_dimmed = "\x1b[2;36m"
    cyan_bold = "\x1b[1;36m"

    grey = "\x1b[37m"
    grey_dimmed = "\x1b[2;37m"
    grey_bold = "\x1b[1:37m"

    default = "\x1b[39m"
    default_dimmed = "\x1b[2;39m"
    default_bold = "\x1b[1;39m"

    reset = "\x1b[0m"

    format_level = "[%(levelname)s]"
    format_date = "%(asctime)s"
    format_name = "%(name)s"
    format_message = "%(asctime)s %(message)s"

    trace_template = (
        f"{grey_bold}[%(levelname)s]{reset} %(asctime)s "
        f":{grey_dimmed}%(name)s{reset}: {grey_dimmed}%(message)s{reset}"
    )

    debug_template = (
        f"{grey_bold}[%(levelname)s]{reset} %(asctime)s "
        f": {grey_dimmed}%(name)s{reset}: {grey_dimmed}%(message)s{reset}"
    )

    info_template = (
        f"{cyan_bold}[%(levelname)s]{reset} %(asctime)s "
        f":{grey_dimmed}%(name)s{reset}: {default_dimmed}%(message)s{reset}"
    )

    warning_template = (
        f"{yellow_bold}[%(levelname)s]{reset} %(asctime)s "
        f":{grey_dimmed}%(name)s{reset}: {yellow_dimmed}%(message)s{reset}"
    )

    error_template = (
        f"{magenta_bold}[%(levelname)s]{reset} %(asctime)s "
        f":{grey_dimmed}%(name)s{reset}: {magenta_dimmed}%(message)s{reset}"
    )

    critical_template = (
        f"{red_bold}[%(levelname)s]{reset} %(asctime)s "
        f":{grey_dimmed}%(name)s{reset}: {red_dimmed}%(message)s{reset}"
    )

    FORMATS = {
        5: trace_template,
        logging.DEBUG: debug_template,
        logging.INFO: info_template,
        logging.WARNING: warning_template,
        logging.ERROR: error_template,
        logging.CRITICAL: magenta_bold + format_level + reset + format_message,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


# logger = logging.getLogger("payboy")
# sh = logging.StreamHandler()
# sh.setFormatter(CustomFormatter())
# logger.setLevel(10)
# logger.addHandler(sh)


class CustomJSONFormatter(logging.Formatter):
    def __init__(self, fmt):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        logging.Formatter.format(self, record)
        return json.dumps(self.get_log(record), indent=2)

    def get_log(self, record):
        d = {
            "time": record.asctime,
            "process_name": record.processName,
            "process_id": record.process,
            "thread_name": record.threadName,
            "thread_id": record.thread,
            "level": record.levelname,
            "logger_name": record.name,
            "pathname": record.pathname,
            "line": record.lineno,
            "message": record.message,
        }

        if hasattr(record, "extra_info"):
            d["req"] = record.extra_info["req"]
            d["res"] = record.extra_info["res"]
        return d



import datetime
import math
import time

import sqlalchemy.types as sqltypes

from pathlib import Path
from typing import (Annotated, Any, Coroutine, Dict, Generic, List, Sequence,
                    Tuple, Type, TypeVar, Union, Optional, TYPE_CHECKING, Mapping, OrderedDict)
from sqlalchemy import exc
from sqlalchemy import (BinaryExpression, BooleanClauseList, ClauseList,
                        ForeignKey, Integer, Column, ColumnClause, Cast,Row, Select, String,
                        TypeDecorator, UnaryExpression, Enum as SQLEnum,)
from sqlalchemy import BigInteger, Integer, DateTime, Cast, Uuid, Boolean
from sqlalchemy.ext.asyncio import AsyncAttrs, AsyncEngine, AsyncSession
from sqlalchemy.orm import (ColumnProperty, DeclarativeBase,
                            InstrumentedAttribute, Mapped)
from sqlalchemy.orm.relationships import Relationship
from sqlalchemy.exc import IllegalStateChangeError, AmbiguousForeignKeysError, ArgumentError, AwaitRequired, SQLAlchemyError, StatementError, SAWarning, NotSupportedError, DatabaseError, DBAPIError, IntegrityError, UnboundExecutionError
ClausesType = TypeVar(
    "ClausesType",
    ClauseList,
    BinaryExpression,
    BooleanClauseList,
    UnaryExpression,
    bool,
)

DBModel = TypeVar("DBModel", bound="Base")


class TimeStamp(TypeDecorator):
    """Prefixes Unicode values with "PREFIX:" on the way in and
    strips it off on the way out.
    """

    impl = Integer
    cache_ok = True
    _current_timestamp_length: int = math.ceil(math.log10(int(time.time())))

    def process_bind_param(self, value, dialect):
        # print("process bind param", value, dialect)
        _value  = None
        if isinstance(value, datetime.datetime):
            _value = int(value.timestamp())
        elif isinstance(value, float):
            _value = int(value)
        elif isinstance(value, str):
            _value = int(value)
        ts = _value
        while True:
            a = math.ceil(math.log10(_value))
            if a > self._current_timestamp_length:
                _value //= 10
                continue
            elif a < self._current_timestamp_length:
                _value *= 10
                continue
            break
        return _value

    def process_result_value(self, value, dialect):
        # print("process result value", value, dialect)
        _value = None
        if isinstance(value, datetime.datetime):
            _value = int(value.timestamp())
        elif isinstance(value, float):
            _value = int(value)
        elif isinstance(value, str):
            _value = int(
                datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S").timestamp()
            )
        while True:
            a = math.ceil(math.log10(_value))
            if a > self._current_timestamp_length:
                _value //= 10
                continue
            elif a < self._current_timestamp_length:
                _value *= 10
                continue
            break
        return _value


class DictType(TypeDecorator):
    import json as _json

    # convert unix timestamp to datetime object
    impl = String

    # convert datetime object to unix timestamp when inserting data to database
    def process_bind_param(self, value, dialect) -> str:
        if value is not None:
            return self._json.dumps(value)
        else:
            return value

    def process_result_value(self, value, dialect) -> dict:
        if value is not None:
            value = self._json.loads(value)
        return value


class FilterParams:
    limit: int = None
    offset: int = None
    order_by: Union[str, InstrumentedAttribute] = None
    group_by: Union[str, InstrumentedAttribute] = None
    search: str = None

    def __init__(
            self,
            order_by: str = None,
            group_by: str = None,
            limit: int = None,
            offset: int = None,
            search: str = None,
    ):
        self.limit = limit
        self.offset = offset
        self._order_by = order_by
        self._group_by = group_by
        self._search = search

    def __call__(self, model: DBModel):
        """hack to be able to use getattr on the model"""
        if self._order_by:
            self.order_by: InstrumentedAttribute = getattr(model, self._order_by)
        if self._group_by:
            self.group_by: InstrumentedAttribute = getattr(model, self._group_by)
        return self


class ResultWithStatsType(Generic[DBModel]):
    data: Union[List[DBModel], DBModel]
    offset: int
    limit: int
    total_pages: int
    num_pages: int
    cur_page: int
    count: int
    range: Tuple[int, int]
    stats: Dict[str, int]





import asyncio
import logging
import pathlib
import secrets
from contextlib import asynccontextmanager

import fastapi
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware

from ._router import BaseRouter
from ._repository import BaseRepository


@asynccontextmanager
async def lifespan(fastapi: FastAPI):
    yield


@property
def root_router(app: FastAPI):
    return app.router  # noqa

# monkeypatching since app.router is somehow hidden
FastAPI.root_router = root_router  # noqa

app: FastAPI = FastAPI(
    debug=True,
    lifespan=lifespan,
    redirect_slashes=True,
    swagger_ui_parameters={"defaultModelsExpandDepth": -1},
    openapi_url="/api/openapi.json",
)

app.add_middleware(
    CORSMiddleware,  # noqa
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

def _generate_key() -> str:
    return secrets.token_urlsafe(16)


app.add_middleware(SessionMiddleware, secret_key=_generate_key())


logging.getLogger("uvicorn.error").name = "uvicorn"



from __future__ import annotations
import logging
from typing import Generic, List, Type, TypeVar, Union, Optional

from fastapi import HTTPException


from .types import schema_model_conf, BaseModel
from .types import ReadModel, CreateModel, UpdateModel
from .types import DBModel, FilterParams, ResultWithStatsType

logger = logging.getLogger(__name__)


class BaseRepository(Generic[DBModel, ReadModel, CreateModel, UpdateModel]):
    """
    A base class a "repository", in which database operations simply can be defined,
    by passing in a DB model (SQLAlchemy), and Pydantic Models for Read, Create, Update

    the returned object consists of 4 main actions:

        - list()
            returns a list of items, based on filter params
        - get()
            returns a single item, based on id
        - create()
            creates a new item
        - update()
            updates an existing item
        - delete()
            deletes an item based on id

    """
    def __init__(self, model: Type[DBModel]):
        self.model = model
        super().__init__()

    async def list(
            self, filter_params: FilterParams
    ) -> ResultWithStatsType[ReadModel]:
        results: ResultWithStatsType[ReadModel] = await self.model.query_with_stats(
            offset=filter_params.offset,
            limit=filter_params.limit,
            order_by=filter_params(self.model).order_by,
            group_by=filter_params(self.model).group_by,
            search=filter_params.search,
        )
        return results

    async def get(
            self, item_id: str = None, search: str = None,
    ) -> ReadModel:
        item = None
        if item_id:
            item: DBModel = await self.model.query_single(self.model.id == item_id)
        elif search:
            item: DBModel = await self.model.query_single(search=search)
        if not item:
            raise Exception("Item not found with id '%s'" % item_id)
            # raise HTTPException(404, "could not find item with id '%s'" % item_id)
        return item

    async def update(
            self, item_id: str, data: UpdateModel,
    ) -> ReadModel:
        item: DBModel = await self.model.query_single(self.model.id == item_id)
        for k, v in dict(data).items():
            if not hasattr(item, k):
                logger.warning(
                    "during patching of %s, a key was not found in the to-be-patched object: %s"
                    % (k, item)
                )
                continue
            if getattr(item, k) != v:
                setattr(item, k, v)
        try:
            await item.update()
            return await self.model.query_single(self.model.id == item.id)

        except:  # noqa
            # logger.exception("could not update %s" % item, exc_info=True)
            raise Exception("could not update item with id '%s' and data '%s" % (item_id, data))

    async def delete(self, item_id: str) -> str:
        item: DBModel = await self.model.query_single(self.model.id == item_id)
        try:
            await item.delete()
            return item.id
        except:  # noqa
            raise Exception("could not delete item with id: %s  -> %s" % (item_id, item))
            # raise HTTPException(405, "could not delete item with id '%s'" % item_id)

    async def create(self, data: CreateModel) -> ReadModel:
        item: DBModel = self.model()
        for k, v in dict(data).items():
            if not hasattr(item, k):
                logger.warning(
                    "during creation of %s, a key was not found in the to-be-patched object: %s"
                    % (k, item)
                )
                continue
            setattr(item, k, v)

        # if item._get_relationship_type("user") == User:  # noqa
        #     setattr(item, "user", user)

        try:
            await item.update()
            # this is needed to avoid DetachedstateError for .user attribute
            return await self.model.query_single(self.model.id == item.id)
        except:  # n oqa
            raise Exception("could not create item fromdata '%s'" % (data))





__all__ = [
    'app', 'Base', 'DBModel',
    'ReadModel', 'UpdateModel',
    'CreateModel', 'BaseRouter',
    'BaseRepository', 'db', 'web',
    'webtypes', 'dbtypes',
    'Base',
    'PaginatedApiResponse',
    'Pagination',
    'HTTPException',
    'factory','ResultWithStatsType',
    'configure_factory',
    'configure_logging']

from fastapis.db import types as dbtypes
from fastapis.db.base import Base, factory, configure_factory
from fastapis.db.types import *
from fastapis.web import BaseRepository
from fastapis.web import BaseRouter
from fastapis.web import app
from fastapis.web import types as webtypes
from fastapis.web.types import *
from .common._config import configure_logging

__version__ = "0.13"



from typing import TypeVar, Optional, Generic, List

from fastapi import Request, File, Form, UploadFile, Query, Path, Body, Cookie, HTTPException, WebSocketException, WebSocket, WebSocketDisconnect, Header
from fastapi.responses import *
from fastapi.requests import Request
from fastapi.security.oauth2 import *
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.middleware import Middleware

from pydantic import *
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel, to_snake

from ..db.types import FilterParams, DBModel, ResultWithStatsType


M = TypeVar("M", bound=BaseModel)


schema_model_conf = ConfigDict(
    alias_generator=to_camel,
    from_attributes=True,
    populate_by_name=True,
    use_enum_values=True,
)

ReadModel = TypeVar("ReadModel", bound=BaseModel)
UpdateModel = TypeVar("UpdateModel", bound=BaseModel)
CreateModel = TypeVar("CreateModel", bound=BaseModel)


class Pagination(BaseModel):
    model_config = schema_model_conf
    count_max: Optional[int] = None
    per_page: Optional[int] = None
    count: Optional[int] = None
    page_max: Optional[int] = None
    has_next: Optional[bool] = None
    has_prev: Optional[bool] = None
    page_index: Optional[int] = None
    range: Optional[tuple[int, int]] = None


class PaginatedApiResponse(Pagination, Generic[M]):
    model_config = schema_model_conf
    data: Optional[List[M]] = None






from __future__ import annotations

import asyncio
import logging
import uuid
import warnings
from asyncio import Task
from typing import Any, Coroutine, Dict, Generic, List, Sequence, Type

import sqlalchemy.exc
import sqlalchemy.ext
import sqlalchemy.orm.collections
from sqlalchemy import Select, all_, and_, any_, func, or_, select
from sqlalchemy.orm import mapped_column, relationship, Mapped, MappedColumn, ColumnProperty
from sqlalchemy.ext.asyncio import async_sessionmaker,create_async_engine, create_async_pool_from_url, async_engine_from_config, async_scoped_session, async_session, async_scoped_session, async_object_session, async_engine_from_config, close_all_sessions, create_async_pool_from_url

from .types import (Annotated, AsyncAttrs, AsyncEngine, AsyncSession,
                    DBModel, BinaryExpression, BooleanClauseList, ClauseList,
                    ClausesType, ColumnProperty, Coroutine, DeclarativeBase,
                    Dict, ForeignKey,
                    InstrumentedAttribute, Integer, List, Mapped,
                    Relationship, ResultWithStatsType, Row, Select, Sequence,
                    String, TimeStamp, Tuple, Type, TypeDecorator, TypeVar,
                    UnaryExpression)
from .types import IllegalStateChangeError, AmbiguousForeignKeysError, ArgumentError, AwaitRequired, SQLAlchemyError, StatementError, SAWarning, NotSupportedError, DatabaseError, DBAPIError, IntegrityError, UnboundExecutionError
from .types import DictType, TimeStamp
logger = logging.getLogger(__name__)


class Factory:
    """
    Factory for database related stuff
    holder of the engine (AsyncEngine), session_maker[AsyncSession]
    can be used as async contextmanager to yield a sqlalchemy AsyncSession
    also responsible for creating the db schema (create_schema())
    """

    _engine: AsyncEngine = None
    _session_maker: async_sessionmaker[AsyncSession] = None
    _session: AsyncSession | None = None
    _scheme_created = False
    connection_string: str = None
    echo_queries: bool = None

    @classmethod
    def engine(cls):
        if not cls._engine:
            cls._engine = create_async_engine(
                cls.connection_string,
                echo=cls.echo_queries
            )
        return cls._engine

    @classmethod
    def session(cls) -> AsyncSession:
        if not cls._session_maker:
            cls._session_maker = async_sessionmaker[AsyncSession](
                bind=cls.engine(), expire_on_commit=False,  # close_resets_only=False
            )
        if not cls._session:
            cls._session = cls._session_maker()
        return cls._session

    @classmethod
    def create_schema(cls) -> Coroutine | Task | None:
        """
        creates the database and/or schema
        this can be awaited (async) or called (sync)

        :return:
        """
        logger.debug("creating schema")

        async def wrapped():
            engine = cls.engine()
            # async with cls() as s:
            #     connection = await s.connection()
            #     engine = connection.engine
            async with engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)
                # cls.engine().begin() as conn:

        try:
            # check if we're in an async context or not
            loop = asyncio.get_running_loop()
            return loop.create_task(wrapped())
        except RuntimeError:
            # we're in a sync context, call in help from asyncio.run
            return asyncio.run(wrapped())
        finally:
            cls._scheme_created = True

    @classmethod
    async def __aenter__(cls) -> AsyncSession:
        if not cls._scheme_created:
            await cls.create_schema()
        cls._current_session = cls.session()
        return cls._current_session

    @classmethod
    async def __aexit__(cls, exc_type, exc_val, exc_tb) -> None:
        try:
            if cls._current_session.new or cls._current_session.dirty or cls._current_session.deleted:
                await cls._current_session.commit()
                return await cls._current_session.close()
        except sqlalchemy.exc.InvalidRequestError as e:
            logger.warning("invalidrequesterror when on commit:\n%s" % e.args)
            raise
        except:  # noqa
            await asyncio.shield(cls._current_session.rollback())

            raise
        finally:
            try:
                # None
                await asyncio.shield(cls._current_session.close())
                # await cls._session.close()
                pass
            except IllegalStateChangeError as e:
                logger.warning("illegalstate error while closing db in context manager:\n%s" % e.args)
                pass
            finally:
                cls._session = None


def factory(*, connection_string: str = "sqlite+aiosqlite:///database.db", echo_queries: bool = False) -> Factory:
    if Factory.connection_string is None:
        warnings.warn(f"""\n
                    connection_string is not set on Factory.
                    using default value now: '{connection_string}'
                    to set your own values, run factory() for the first time using keywords,
                    eg: factory(connection_string="some conn string", echo_queries=True )  
                    """, category=UserWarning)
        Factory.connection_string = connection_string
        # raise RuntimeError("connection_string is not set on Factory. run configure_factory first")
    if Factory.echo_queries is None:
        warnings.warn(f"""\n
                            echo_queries is not set on Factory.
                            using default value now: '{echo_queries}'
                            to set your own values, run factory() for the first time using keywords,
                            eg: factory(connection_string="some conn string", echo_queries=True )  
                            """, category=UserWarning)
        Factory.echo_queries = echo_queries
        # raise RuntimeError("echo_queries is not set on Factory.  run configure_factory first")
    return Factory()


def configure_factory(connection_string: str = "sqlite+aiosqlite:///database.db",
                      echo_queries: bool = False):
    Factory.connection_string = connection_string
    Factory.echo_queries = echo_queries
    return Factory()


class Base(AsyncAttrs, DeclarativeBase):
    id_: Mapped[int] = mapped_column(primary_key=True, unique=True, autoincrement=True)
    id: Mapped[str] = mapped_column(default=lambda: str(uuid.uuid4()))

    updated_at: Mapped[int] = mapped_column(
        TimeStamp, default=func.now(), onupdate=func.now()
    )

    created_at: Mapped[int] = mapped_column(TimeStamp, server_default=func.now())


    def __init__(self, *args, **kwargs):
        super().__init__()
        for k, v in dict(*args, **kwargs).items():
            if not hasattr(self, k):
                logger.warning(
                    f"trying to assign %s to %s, but %s has no attribute %s"
                    % (k, self, self, k)
                )
            setattr(self, k, v)

    @property
    def dict(self) -> Dict[str, Any]:
        attrs = dict(self.__mapper__.attrs)  # noqa
        ret = {}
        for k, v in attrs.items():
            if hasattr(v, "columns"):
                # we have regular column
                value = getattr(self, k)
                match type(value).__name__:
                    case "str":
                        ret[k] = value
                    case "int":
                        ret[k] = value
                    case "float":
                        ret[k] = value
                    case _:
                        ret[k] = str(value)
            elif hasattr(v, "argument"):
                # we have a relation
                try:
                    val = getattr(self, k)
                    if isinstance(
                            val,
                            (
                                    sqlalchemy.orm.collections.InstrumentedList,
                                    sqlalchemy.orm.collections.InstrumentedSet,
                            ),
                    ):
                        val = [v.as_dict() for v in val]
                    elif isinstance(val, Base):
                        val = val.dict
                    ret[k] = val
                except:  # noqa
                    raise
        return ret

    @staticmethod
    def __make_string_search_query__(
            cls: Type[DBModel],
            search: str,
            q: Select,
    ):
        string_columns = [
            column
            for column in cls.__mapper__.c
            if "string" in type(column.type).__name__.lower()
        ]
        # print('string columns', string_columns)
        matching = [column.ilike(f"%{search}%") for column in string_columns]
        # print(matching)
        return q.filter(or_(*matching))

    @classmethod
    def _get_relationship_type(cls, field: str) -> Type | None:
        if not hasattr(cls, field):
            return None
        attrs = dict(cls.__mapper__.attrs)  # noqa
        val = attrs.get(field)
        if not val:
            return None
        if isinstance(val, Relationship):
            return val.mapper.class_

    @staticmethod
    def __make_select__(
            cls: Type[DBModel],
            *clauses: ClausesType,
            order_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            group_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            limit: int = None,
            offset: int = None,
    ):
        q = select(cls)
        if clauses and clauses[0] is not None:
            q = q.where(*clauses)
        if order_by is not None:
            q = q.order_by(order_by)
        if group_by:
            q = q.group_by(group_by)
        if limit:
            q = q.limit(limit)
        if offset:
            if not limit:
                limit = 16
            if not isinstance(limit, int):
                limit = int(limit)
            q = q.offset(offset * limit)
        return q

    @classmethod
    async def query(
            cls: Type[DBModel],
            *clauses: ClausesType,
            order_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            group_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            limit: int = None,
            offset: int = None,
            search: str = None,
    ) -> List[DBModel]:
        """
        query object from the database, it will always return a list
        :param clauses:
        :param order_by:
        :param group_by:
        :param limit:
        :param offset:
        :param search:
        :return:
        """

        q = cls.__make_select__(
            cls,
            *clauses,
            order_by=order_by,
            group_by=group_by,
            limit=limit,
            offset=offset,
        )
        if search:
            q = cls.__make_string_search_query__(cls, search, q)
        q = q.order_by(cls.id_.desc())

        async with factory() as session:
            scalars = await session.scalars(q)
            return list(scalars.all())

    @classmethod
    async def query_with_stats(
            cls: Type[DBModel],
            *clauses: ClausesType,
            order_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            group_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            limit: int = 16,
            offset: int = 0,
            search: str = None,
    ) -> ResultWithStatsType[DBModel]:
        """
        query object from the database, it will always return a list
        :param clauses:
        :param order_by:
        :param group_by:
        :param limit:
        :param offset:
        :param search:
        :return:
        """
        q = select(cls, func.count(cls.id_).over().label('count'))
        if clauses and clauses[0] is not None:
            q = q.where(*clauses)
        if order_by is not None:
            q = q.order_by(order_by)
        if group_by:
            q = q.group_by(group_by)
        if limit:
            q = q.limit(limit)
        if offset:
            if not limit:
                limit = 16
            if not isinstance(limit, int):
                limit = int(limit)
            q = q.offset(offset * limit)
        if search:
            q = cls.__make_string_search_query__(cls, search, q)
        q = q.order_by(cls.id_.desc())
        async with factory() as session:
            result = await session.execute(q)
            results = result.fetchall()
            # pagination = Pagination()
            #
            # pagination_ = [r[0] for r in list(results)]
            # try:
            #     count_max_ = results[0][-1]
            # except IndexError:
            #     count_max_ = 0
            #
            # offset_ = getattr(q, "_offset") or 0
            # limit_ = getattr(q, "_limit") or 16
            #
            # total_pages_, total_remainder_ = divmod(count_max_, limit_)
            # if total_remainder_ and count_max_ // limit_:
            #     total_pages_ += 1
            # cur_page_, cur_remainder_ = divmod(offset_, limit_)
            # if cur_remainder_ and offset_ // limit_:
            #     cur_page_ += 1
            # num_pages_ = total_pages_
            # current_page_ = cur_page_
            # count_ = min(len(pagination_), limit_)
            # range_ = offset_, offset_ + min(count_, limit_)
            #
            # pagination.offset = offset_
            # pagination.count = count_
            # pagination.range = range_
            # pagination.total_pages = total_pages_
            # pagination.num_pages = num_pages_
            # pagination.limit = limit_
            # pagination.cur_page = cur_page_
            #
            # for r in pagination_:
            #     r.pagination = pagination
            # return pagination_

            return ResultWithStats(rows=results, select_instance=q)

    @classmethod
    async def query_single(
            cls: Type[DBModel],
            *clauses: ClausesType,
            order_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            group_by: UnaryExpression | BinaryExpression | InstrumentedAttribute = None,
            limit: int = None,
            offset: int = None,
            search: str = None,
    ) -> DBModel:
        """
        query a single object from the database, returns a single item
        :param clauses:
        :param order_by:
        :param group_by:
        :param limit:
        :param offset:
        :param search:
        :return:
        """
        q = cls.__make_select__(
            cls,
            *clauses,
            order_by=order_by,
            group_by=group_by,
            limit=limit,
            offset=offset,
        )
        if search:
            q = cls.__make_string_search_query__(cls, search, q)
        async with factory() as session:
            scalars = await session.scalars(q)
            return scalars.first()

    @classmethod
    async def count(cls: DBModel) -> int:
        async with factory() as session:
            scalars = await session.scalars(select(func.count(cls.id_)))
            return scalars.first()

    async def update(self):
        """
        updates or adds a object to/in the database
        :return:
        """
        async with factory() as session:
            session.add(self)

    async def delete(self):
        """
        delete the object from the database
        :return:
        """
        async with factory() as session:
            await session.delete(self)

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        # if Factory.connection_string is None:
        #     raise RuntimeError("connection_string is not set on Factory. run configure_factory first")
        # if Factory.echo_queries is None:
        #     raise RuntimeError("echo_queries is not set on Factory.  run configure_factory first")
        super().__init_subclass__(*args, **kwargs)
        print('baseclass of Base is created', args, kwargs)


class ResultWithStats(ResultWithStatsType):
    data: DBModel

    def __init__(self, rows: Sequence[Row], select_instance: Select):

        _rows = list(rows)
        self.data = [r[0] for r in _rows]
        try:
            self.count_max = _rows[0][-1]
        except IndexError:
            self.count_max = 0

        self.offset = getattr(select_instance, "_offset") or 0
        self.limit = getattr(select_instance, "_limit") or 16

        total_pages, total_remainder = divmod(self.count_max, self.limit)
        if total_remainder and self.count_max // self.limit:
            total_pages += 1
        cur_page, cur_remainder = divmod(self.offset, self.limit)
        if cur_remainder and self.offset // self.limit:
            cur_page += 1
        self.num_pages = total_pages
        self.current_page = cur_page
        self.count = min(len(self.data), self.limit)
        self.range = self.offset, self.offset + min(len(self.data), self.limit)

    @property
    def stats(self) -> Dict[str, int]:
        s = dict(
            count_max=self.count_max,
            per_page=self.limit,
            count=self.count,
            page_max=self.num_pages,
            has_next=self.range[-1] < self.count_max,
            has_prev=self.range[0] > 0,
            page_index=self.current_page,
            range=self.range
        )
        return s

    def __repr__(self):
        base = f"<ResultWithStats[{{}}]\n\nstats:[{{}}]>"
        return base.format(self.data, self.stats)



import re

from setuptools import setup

setup(
    name="fastapis",
    version=re.search("(?smi)^__version__.+?([0-9.]+)", open("fastapis/__init__.py", encoding="utf-8").read())[1],
    packages=["fastapis"],
    install_requires=[
        "aiomysql==0.2.0",
        "aiosqlite==0.21.0",
        "annotated-types==0.7.0",
        "anyio==4.9.0",
        "asyncpg==0.30.0",
        "click==8.2.1",
        "colorama==0.4.6",
        "fastapi==0.115.12",
        "greenlet==3.2.2",
        "h11==0.16.0",
        "idna==3.10",
        "itsdangerous==2.2.0",
        "pydantic==2.11.5",
        "pydantic_core==2.33.2",
        "PyMySQL==1.1.1",
        "python-multipart==0.0.20",
        "PyYAML==6.0.2",
        "sniffio==1.3.1",
        "SQLAlchemy==2.0.41",
        "starlette==0.46.2",
        "typing-inspection==0.4.1",
        "typing_extensions==4.13.2",
        "uvicorn==0.34.2",
    ],
    long_description_content_type="text/markdown",
    url="https://github.com/ultrafunkamsterdam/fastapis",
    license="GPL-3.0",
    author="UltrafunkAmsterdam",
    author_email="info@blackhat-security.nl",
    description="SQLAlchemy with Fastapi asyncio for humans",

)
